// KeyMapDlg.cpp : implementation file
//

#include "stdafx.h"
#include "midifish.h"
#include "KeyMapDlg.h"
#include "keymap.h"
#include "doc.h"
#include "midi.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif



#define SPLIT_GAP	16



KeyMapDlg::KeyMapDlg(KeyMap *map, Doc *doc)
	: CDialog(KeyMapDlg::IDD, NULL),
	  mZones(map, this)
{
	mMap = map;
	mDoc = doc;

	mEntryX = 0;
	mEntryY = 0;

	mEntry = &map->mMap[mEntryY][mEntryX];

	mDragX = -1;
	mDragY = -1;

	mDragCol = -1;
	mDragRow = -1;

	mDragging = false;

	//{{AFX_DATA_INIT(KeyMapDlg)
		// NOTE: the ClassWizard will add member initialization here
	//}}AFX_DATA_INIT
}



void KeyMapDlg::DoDataExchange(CDataExchange* pDX)
{
	int i;

	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(KeyMapDlg)
	DDX_Control(pDX, IDC_ZONES, mZones);
	DDX_Control(pDX, IDC_VEL_OFFSET_LABEL, mVelOffsetLabel);
	DDX_Control(pDX, IDC_VEL_SCALE_LABEL, mVelScaleLabel);
	DDX_Control(pDX, IDC_VSPLIT_CROSSFADE, mVCrossfade);
	DDX_Control(pDX, IDC_VSPLIT_CONTROLLER, mVController);
	DDX_Control(pDX, IDC_VEL_SCALE_SLIDER, mVelScaleSlider);
	DDX_Control(pDX, IDC_VEL_SCALE_EDIT, mVelScaleEdit);
	DDX_Control(pDX, IDC_VEL_OFFSET_SLIDER, mVelOffsetSlider);
	DDX_Control(pDX, IDC_VEL_OFFSET_EDIT, mVelOffsetEdit);
	DDX_Control(pDX, IDC_PORT_OUT, mPortOut);
	DDX_Control(pDX, IDC_KEY_OUT, mKeyOut);
	DDX_Control(pDX, IDC_KEY_IN, mKeyIn);
	DDX_Control(pDX, IDC_HSPLIT_CROSSFADE, mHCrossfade);
	DDX_Control(pDX, IDC_HSPLIT_CONTROLLER, mHController);
	//}}AFX_DATA_MAP

	if (!pDX->m_bSaveAndValidate) {
		mKeyIn.SetRedraw(false);
		mKeyIn.ResetContent();
		mKeyIn.AddString(mMap->get_name(mMap->mKey).c_str());
		mKeyIn.SetRedraw(true);

		// output note combo box
		mKeyOut.SetRedraw(false);
		mKeyOut.ResetContent();

		mKeyOut.AddString("<none>");

		for (i=0; i<128; i++)
			mKeyOut.AddString(mMap->get_name(i).c_str());

		mKeyOut.SetCurSel(mEntry->mKey+1);

		mKeyOut.SetRedraw(true);
		mKeyOut.Invalidate();

		// MIDI channel combo box
		mPortOut.SetRedraw(false);
		mPortOut.ResetContent();

		for (i=0; i<Doc::NUM_OUTS; i++) {
			MidiOutput *midi = mDoc->mOut[i];

			if (midi->mPort < midi->mInterfaces.size()) {
				CString s;
				s.Format("%s [%d]", midi->mInterfaces[midi->mPort].c_str(), midi->mChannel+1);
				mPortOut.AddString(s);
			}
			else
				mPortOut.AddString("<none>");
		}

		mPortOut.SetCurSel(mEntry->mPort);

		mPortOut.SetRedraw(true);
		mPortOut.Invalidate();

		// velocity range and scale
		mVelScaleSlider.SetRange(0, 512, true);
		mVelScaleSlider.SetPos(mEntry->mVelScale);

		mVelOffsetSlider.SetRange(-128, 128, true);
		mVelOffsetSlider.SetPos(mEntry->mVelOffset);

		CString s;

		s.Format("%.4f", (float)mEntry->mVelScale / 256.0);

		if (s.Find('.') >= 0) {
			while (s.Right(1) == "0")
				s = s.Left(s.GetLength()-1);

			if (s.Right(1) == ".")
				s = s.Left(s.GetLength()-1);
		}


		mVelScaleEdit.SetWindowText(s);

		s.Format("%d", mEntry->mVelOffset);
		mVelOffsetEdit.SetWindowText(s);

		for (i=0; i<2; i++) {
			// split controller
			CComboBox &combo = (i) ? mHController : mVController;

			combo.SetRedraw(false);
			combo.ResetContent();

			for (int j=0; j<CtrlMap::NUM_TYPES; j++)
				combo.AddString(CtrlMap::mNames[j]);

			combo.SetCurSel(mMap->mSplitController[i]);

			combo.SetRedraw(true);

			// split crossfade
			CSliderCtrl &slider = (i) ? mHCrossfade : mVCrossfade;

			slider.SetRange(0, 256);
			slider.SetPos(mMap->mSplitCrossfade[i]);
		}

		// hide if we are disabled
		mPortOut.ShowWindow(mEntry->mKey >= 0);

		mVelScaleSlider.ShowWindow(mEntry->mKey >= 0);
		mVelOffsetSlider.ShowWindow(mEntry->mKey >= 0);

		mVelScaleEdit.ShowWindow(mEntry->mKey >= 0);
		mVelOffsetEdit.ShowWindow(mEntry->mKey >= 0);

		mVelScaleLabel.ShowWindow(mEntry->mKey >= 0);
		mVelOffsetLabel.ShowWindow(mEntry->mKey >= 0);

		// deal with split changes
		mZones.Invalidate(false);

		GetDlgItem(IDC_HSPLIT_ADD)->EnableWindow(mMap->mSplitPoint[1][KeyMap::MAX_SPLITS-1] >= 256);
		GetDlgItem(IDC_HSPLIT_REMOVE)->EnableWindow(mMap->mSplitPoint[1][1] < 256);

		GetDlgItem(IDC_VSPLIT_ADD)->EnableWindow(mMap->mSplitPoint[0][KeyMap::MAX_SPLITS-1] >= 256);
		GetDlgItem(IDC_VSPLIT_REMOVE)->EnableWindow(mMap->mSplitPoint[0][1] < 256);
	}
}



BEGIN_MESSAGE_MAP(KeyMapDlg, CDialog)
	//{{AFX_MSG_MAP(KeyMapDlg)
	ON_BN_CLICKED(IDC_HSPLIT_ADD, OnHsplitAdd)
	ON_BN_CLICKED(IDC_VSPLIT_ADD, OnVsplitAdd)
	ON_CBN_SELCHANGE(IDC_HSPLIT_CONTROLLER, OnSelchangeHsplitController)
	ON_CBN_SELCHANGE(IDC_VSPLIT_CONTROLLER, OnSelchangeVsplitController)
	ON_CBN_SELCHANGE(IDC_KEY_OUT, OnSelchangeKeyOut)
	ON_CBN_SELCHANGE(IDC_PORT_OUT, OnSelchangePortOut)
	ON_EN_CHANGE(IDC_VEL_OFFSET_EDIT, OnChangeVelOffsetEdit)
	ON_EN_CHANGE(IDC_VEL_SCALE_EDIT, OnChangeVelScaleEdit)
	ON_WM_DESTROY()
	ON_WM_HSCROLL()
	ON_WM_TIMER()
	ON_WM_LBUTTONDOWN()
	ON_BN_CLICKED(IDC_HSPLIT_REMOVE, OnHsplitRemove)
	ON_BN_CLICKED(IDC_VSPLIT_REMOVE, OnVsplitRemove)
	ON_WM_LBUTTONUP()
	ON_WM_MOUSEMOVE()
	ON_WM_CAPTURECHANGED()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()



BOOL KeyMapDlg::OnInitDialog() 
{
	CDialog::OnInitDialog();
	
	mTimer = SetTimer(3, 50, NULL);

	return TRUE;
}



void KeyMapDlg::OnDestroy() 
{
	KillTimer(mTimer);

	CDialog::OnDestroy();
}



void KeyMapDlg::OnSelchangeKeyOut() 
{
	mEntry->mKey = mKeyOut.GetCurSel() - 1;

	UpdateData(false);
}



void KeyMapDlg::OnSelchangePortOut() 
{
	mEntry->mPort = mPortOut.GetCurSel();
}



void KeyMapDlg::OnSelchangeHsplitController() 
{
	mMap->mSplitController[1] = mHController.GetCurSel();
}



void KeyMapDlg::OnSelchangeVsplitController() 
{
	mMap->mSplitController[0] = mVController.GetCurSel();
}



void KeyMapDlg::OnChangeVelOffsetEdit() 
{
	CString s;

	mVelOffsetEdit.GetWindowText(s);

	int n = atoi(s);

	if (n < -128)
		n = -128;
	else if (n > 128)
		n = 128;

	mEntry->mVelOffset = n;

	mVelOffsetSlider.SetPos(n);
}



void KeyMapDlg::OnChangeVelScaleEdit() 
{
	CString s;

	mVelScaleEdit.GetWindowText(s);

	int n = atof(s) * 256.0;

	if (n < 0)
		n = 0;
	else if (n > 512)
		n = 512;

	mEntry->mVelScale = n;

	mVelScaleSlider.SetPos(n);
}



static void slide(int &val, CSliderCtrl &slider, CEdit &edit, bool floating_point)
{
	val = slider.GetPos();

	CString s;

	if (floating_point) {
		s.Format("%.4f", (float)val / 256.0);

		if (s.Find('.') >= 0) {
			while (s.Right(1) == "0")
				s = s.Left(s.GetLength()-1);

			if (s.Right(1) == ".")
				s = s.Left(s.GetLength()-1);
		}
	}
	else
		s.Format("%d", val);

	edit.SetWindowText(s);
}



void KeyMapDlg::OnHScroll(UINT nSBCode, UINT nPos, CScrollBar *pScrollBar) 
{
	CSliderCtrl *slider = (CSliderCtrl *)pScrollBar;

	if (slider == &mVCrossfade)
		mMap->mSplitCrossfade[0] = mVCrossfade.GetPos();
	else if (slider == &mHCrossfade)
		mMap->mSplitCrossfade[1] = mHCrossfade.GetPos();
	else if (slider == &mVelScaleSlider)
		slide(mEntry->mVelScale, mVelScaleSlider, mVelScaleEdit, true);
	else if (slider == &mVelOffsetSlider)
		slide(mEntry->mVelOffset, mVelOffsetSlider, mVelOffsetEdit, false);
}



static void add_split(KeyMap *map, int axis)
{
	for (int i=1; i<KeyMap::MAX_SPLITS; i++) {
		if (map->mSplitPoint[axis][i] >= 256) {
			map->mSplitPoint[axis][i] = 256 - SPLIT_GAP;

			while (i > 0) {
				if (map->mSplitPoint[axis][i] > map->mSplitPoint[axis][i+1] - SPLIT_GAP)
					map->mSplitPoint[axis][i] = map->mSplitPoint[axis][i+1] - SPLIT_GAP;

				i--;
			}

			return;
		}
	}
}



void KeyMapDlg::OnHsplitAdd() 
{
	add_split(mMap, 1);

	UpdateData(false);
}



void KeyMapDlg::OnVsplitAdd() 
{
	add_split(mMap, 0);

	UpdateData(false);
}



static void remove_split(KeyMap *map, int axis)
{
	for (int i=KeyMap::MAX_SPLITS-1; i>0; i--) {
		if (map->mSplitPoint[axis][i] < 256) {
			map->mSplitPoint[axis][i] = 256;
			return;
		}
	}
}



void KeyMapDlg::OnHsplitRemove() 
{
	remove_split(mMap, 1);

	mEntryX = mEntryY = 0;

	mEntry = &mMap->mMap[mEntryY][mEntryX];

	UpdateData(false);
}



void KeyMapDlg::OnVsplitRemove() 
{
	remove_split(mMap, 0);

	mEntryX = mEntryY = 0;

	mEntry = &mMap->mMap[mEntryY][mEntryX];

	UpdateData(false);
}



void KeyMapDlg::OnTimer(UINT nIDEvent) 
{
	bool zone_dirty = false;

	// remap note
	if (mDoc->mIn[1]->mRecentNote >= 0) {
		mEntry->mKey = mDoc->mIn[1]->mRecentNote;

		mKeyOut.SetCurSel(mEntry->mKey+1);

		mPortOut.ShowWindow(mEntry->mKey >= 0);

		mVelScaleSlider.ShowWindow(mEntry->mKey >= 0);
		mVelOffsetSlider.ShowWindow(mEntry->mKey >= 0);

		mVelScaleEdit.ShowWindow(mEntry->mKey >= 0);
		mVelOffsetEdit.ShowWindow(mEntry->mKey >= 0);

		mVelScaleLabel.ShowWindow(mEntry->mKey >= 0);
		mVelOffsetLabel.ShowWindow(mEntry->mKey >= 0);

		zone_dirty = true;
	}

	// redraw controller indicators?
	int c[2];

	for (int i=0; i<2; i++) {
		c[i] = mDoc->mCtrlIn[mMap->mSplitController[i]];

		int min, max;

		CtrlMap::get_range(mMap->mSplitController[i], min, max);

		c[i] -= min;
		c[i] *= 255;

		if (min != max)
			c[i] /= (max - min);
	}

	if ((c[0] != mZones.mCtrl[0]) || (c[1] != mZones.mCtrl[1])) {
		mZones.mCtrl[0] = c[0];
		mZones.mCtrl[1] = c[1];

		zone_dirty = true;
	}

	if (zone_dirty)
		mZones.Invalidate(false);

	// clear
	mDoc->mIn[0]->mRecentNote = -1;
	mDoc->mIn[1]->mRecentNote = -1;

	mDoc->mIn[0]->mRecentController = -1;
	mDoc->mIn[1]->mRecentController = -1;

	mDoc->mThruPort = mEntry->mPort;
}



bool KeyMapDlg::point_to_zone(POINT point, int &x, int &y)
{
	RECT rc;

	mZones.GetWindowRect(&rc);

	ClientToScreen(&point);

	x = point.x - rc.left;
	y = 255 - point.y + rc.top;

	return ((x >= 0) && (x < 256) && (y >= 0) && (y < 256));
}



bool KeyMapDlg::hit_test(int pos_x, int pos_y, int &cell_x, int &cell_y)
{
	bool got_x = false;
	bool got_y = false;

	for (int i=0; i<KeyMap::MAX_SPLITS; i++) {
		if ((pos_x >= mMap->mSplitPoint[1][i]) && (pos_x < mMap->mSplitPoint[1][i+1])) {
			cell_x = i;
			got_x = true;
		}

		if ((pos_y >= mMap->mSplitPoint[0][i]) && (pos_y < mMap->mSplitPoint[0][i+1])) {
			cell_y = i;
			got_y = true;
		}
	}

	return (got_x) && (got_y);
}



void KeyMapDlg::OnLButtonDown(UINT nFlags, CPoint point) 
{
	int x, y;

	GotoDlgCtrl(GetDlgItem(IDC_NONE));

	if (point_to_zone(point, x, y)) {
		// clicked on a row or column edge?
		for (int i=1; i<KeyMap::MAX_SPLITS; i++) {
			if ((mMap->mSplitPoint[1][i] < 256) && (ABS(x - mMap->mSplitPoint[1][i]) < 6))
				mDragCol = i;

			if ((mMap->mSplitPoint[0][i] < 256) && (ABS(y - mMap->mSplitPoint[0][i]) < 6))
				mDragRow = i;
		}

		if ((mDragRow > 0) || (mDragCol > 0)) {
			// start dragging the boundary
			mDragStart = point;

			mDragColOrig = (mDragCol > 0) ? mMap->mSplitPoint[1][mDragCol] : 0;
			mDragRowOrig = (mDragRow > 0) ? mMap->mSplitPoint[0][mDragRow] : 0;

			SetCapture();

			if ((mDragCol > 0) && (mDragRow > 0))
				SetCursor(LoadCursor(NULL, IDC_SIZEALL));
			else if (mDragCol > 0)
				SetCursor(LoadCursor(NULL, IDC_SIZEWE));
			else
				SetCursor(LoadCursor(NULL, IDC_SIZENS));

			mDragging = true;

			mZones.Invalidate(false);
		}
		else if (hit_test(x, y, mEntryX, mEntryY)) {
			// select or drag a zone box
			mEntry = &mMap->mMap[mEntryY][mEntryX];

			UpdateData(false);

			SetCapture();
			SetCursor(LoadCursor(NULL, IDC_CROSS));

			mDragX = mEntryX;
			mDragY = mEntryY;

			mDragging = true;
		}
	}
}



void KeyMapDlg::OnLButtonUp(UINT nFlags, CPoint point) 
{
	if (mDragging) {
		mDragging = false;

		ReleaseCapture();

		if ((mDragX >= 0) && (mDragY >= 0)) {
			mMap->mMap[mDragY][mDragX] = mMap->mMap[mEntryY][mEntryX];

			mEntryX = mDragX;
			mEntryY = mDragY;
		}

		mDragX = mDragY = -1;

		mDragCol = -1;
		mDragRow = -1;

		mEntry = &mMap->mMap[mEntryY][mEntryX];

		UpdateData(false);
	}
}



void KeyMapDlg::OnMouseMove(UINT nFlags, CPoint point) 
{
	if (mDragging) {
		if ((mDragRow > 0) || (mDragCol > 0)) {
			// moving a zone boundary
			int dx = point.x - mDragStart.x;
			int dy = point.y - mDragStart.y;

			bool dirty = false;

			if (mDragRow > 0) {
				int n = mDragRowOrig - dy;

				if (n < mMap->mSplitPoint[0][mDragRow-1] + SPLIT_GAP)
					n = mMap->mSplitPoint[0][mDragRow-1] + SPLIT_GAP;

				if (n > mMap->mSplitPoint[0][mDragRow+1] - SPLIT_GAP)
					n = mMap->mSplitPoint[0][mDragRow+1] - SPLIT_GAP;

				if (mMap->mSplitPoint[0][mDragRow] != n) {
					mMap->mSplitPoint[0][mDragRow] = n;
					dirty = true;
				}
			}

			if (mDragCol > 0) {
				int n = mDragColOrig + dx;

				if (n < mMap->mSplitPoint[1][mDragCol-1] + SPLIT_GAP)
					n = mMap->mSplitPoint[1][mDragCol-1] + SPLIT_GAP;

				if (n > mMap->mSplitPoint[1][mDragCol+1] - SPLIT_GAP)
					n = mMap->mSplitPoint[1][mDragCol+1] - SPLIT_GAP;

				if (mMap->mSplitPoint[1][mDragCol] != n) {
					mMap->mSplitPoint[1][mDragCol] = n;
					dirty = true;
				}
			}

			if (dirty)
				mZones.Invalidate(false);
		}
		else {
			// copying zone contents
			int x, y, dx, dy;

			if ((!point_to_zone(point, x, y)) || (!hit_test(x, y, dx, dy)))
				dx = dy = -1;

			if ((dx != mDragX) || (dy != mDragY)) {
				mDragX = dx;
				mDragY = dy;

				mZones.Invalidate(false);
			}
		}
	}	
}



void KeyMapDlg::OnCaptureChanged(CWnd *pWnd) 
{
	if (mDragging) {
		mDragging = false;

		mDragX = mDragY = -1;

		mDragCol = -1;
		mDragRow = -1;

		mZones.Invalidate(false);
	}
}



ZoneButton::ZoneButton(KeyMap *map, KeyMapDlg *dlg)
{
	mMap = map;

	mDlg = dlg;
	
	mCtrl[0] = mCtrl[1] = -1;

	mDC = NULL;
	mBitmap = NULL;
	mPrevBitmap = NULL;
}



ZoneButton::~ZoneButton()
{
	if (mDC) {
		mDC->SelectObject(mPrevBitmap);

		delete mBitmap;
		delete mDC;
	}
}



void ZoneButton::DrawItem(LPDRAWITEMSTRUCT lpDrawItemStruct)
{
	int i, j;

	CDC *dc = CDC::FromHandle(lpDrawItemStruct->hDC);

	if (!mDC) {
		mDC = new CDC;
		mDC->CreateCompatibleDC(dc);

		mBitmap = new CBitmap;
		mBitmap->CreateCompatibleBitmap(dc, 256, 256);

		mPrevBitmap = mDC->SelectObject(mBitmap);
	}

	mDC->SetTextAlign(TA_CENTER);
	mDC->SetBkMode(TRANSPARENT);

	CFont *prev_font = mDC->SelectObject(dc->GetCurrentFont());

	// draw boxes
	for (i=0; i<KeyMap::MAX_SPLITS; i++) {
		if (mMap->mSplitPoint[0][i] >= 256)
			break;

		for (j=0; j<KeyMap::MAX_SPLITS; j++) {
			if (mMap->mSplitPoint[1][j] >= 256)
				break;

			RECT rc = { mMap->mSplitPoint[1][j], 256-mMap->mSplitPoint[0][i+1], mMap->mSplitPoint[1][j+1], 256-mMap->mSplitPoint[0][i] };

			bool sel = (((i == mDlg->mEntryY) && (j == mDlg->mEntryX)) ||
						((i == mDlg->mDragY)  && (j == mDlg->mDragX)));

			mDC->DrawEdge(&rc, EDGE_SUNKEN, BF_RECT | BF_ADJUST);

			mDC->FillSolidRect(&rc, (sel) ? GetSysColor(COLOR_WINDOW) : GetSysColor(COLOR_BTNFACE));

			if (sel) {
				rc.left += 1;
				rc.top += 1;
				rc.right -= 1;
				rc.bottom -= 1;

				mDC->DrawFocusRect(&rc);
			}
		}
	}

	// draw text
	for (i=0; i<KeyMap::MAX_SPLITS; i++) {
		if (mMap->mSplitPoint[0][i] >= 256)
			break;

		for (j=0; j<KeyMap::MAX_SPLITS; j++) {
			if (mMap->mSplitPoint[1][j] >= 256)
				break;

			RECT rc = { mMap->mSplitPoint[1][j], 256-mMap->mSplitPoint[0][i+1], mMap->mSplitPoint[1][j+1], 256-mMap->mSplitPoint[0][i] };

			bool sel = (((i == mDlg->mEntryY) && (j == mDlg->mEntryX)) ||
						((i == mDlg->mDragY)  && (j == mDlg->mDragX)));

			mDC->SetTextColor((sel) ? GetSysColor(COLOR_WINDOWTEXT) : GetSysColor(COLOR_BTNTEXT));

			mDC->TextOut((rc.left+rc.right)/2, (rc.top+rc.bottom)/2-6, mMap->get_name(mMap->mMap[i][j].mKey).c_str());
		}
	}

	// draw controllers
	if (mCtrl[0] >= 0) {
		CPen pen(PS_DOT, 0, GetSysColor(COLOR_HIGHLIGHT));
		
		CPen *old_pen = mDC->SelectObject(&pen);

		int x = MID(1, mCtrl[1], 254);
		int y = 255 - MID(1, mCtrl[0], 254);

		mDC->MoveTo(0,   y);
		mDC->LineTo(256, y);

		mDC->MoveTo(x, 0);
		mDC->LineTo(x, 256);

		mDC->SelectObject(old_pen);
	}

	// draw
	dc->BitBlt(lpDrawItemStruct->rcItem.left, lpDrawItemStruct->rcItem.top, 256, 256, mDC, 0, 0, SRCCOPY);

	mDC->SelectObject(prev_font);
}


