// Controller mapping class

#include "stdafx.h"
#include "ctrlmap.h"
#include "midifish.h"



// construct a null controller map
CtrlMap::CtrlMap(int type)
{
	mType = type;

	mOut[0] = type;
	mOut[1] = NONE;

	for (int i=0; i<NUM_MAPS; i++) {
		get_range(mType, mInMin[i], mInMax[i]);
		get_range(mOut[i], mOutMin[i], mOutMax[i]);
	}
}



// get a string we can show to the user
string CtrlMap::get_desc()
{
	char tmp[256];

	strcpy(tmp, mNames[mType]);
	strcat(tmp, "\t-> ");

	bool done = false;

	if (mOut[0] >= 0) {
		strcat(tmp, mNames[mOut[0]]);
		done = true;
	}

	if (mOut[1] >= 0) {
		if (done)
			strcat(tmp, "  +  ");

		strcat(tmp, mNames[mOut[1]]);
		done = true;
	}

	if (!done)
		strcat(tmp, "<none>");

	return tmp;
}



// query the range of a controller type
void CtrlMap::get_range(int type, int &min, int &max)
{
	if (type == NONE) {
		min = 0;
		max = 0;
	}
	else if (type == PITCH_BEND) {
		min = -0x2000;
		max =  0x1FFF;
	}
	else {
		min = 0;
		max = 127;
	}
}



// read a controller mapping
void CtrlMap::load(FILE *file)
{
	for (int i=0; i<NUM_MAPS; i++) {
		mOut[i] = fgetl(file);

		mInMin[i] = fgetl(file);
		mInMax[i] = fgetl(file);

		mOutMin[i] = fgetl(file);
		mOutMax[i] = fgetl(file);
	}
}



// save a controller mapping
void CtrlMap::save(FILE *file)
{
	for (int i=0; i<NUM_MAPS; i++) {
		fputl(mOut[i], file);

		fputl(mInMin[i], file);
		fputl(mInMax[i], file);

		fputl(mOutMin[i], file);
		fputl(mOutMax[i], file);
	}
}



const char *CtrlMap::mNames[] =
{
	"Key",
	"Velocity",
	"Key Pressure",
	"Aftertouch",
	"Pitch Bend",
	"Bank Sel MSB",
	"Modulation",
	"Breath",
	"Controller 3",
	"Foot",
	"Portamento",
	"Data Ent MSB",
	"Volume",
	"Balance",
	"Controller 9",
	"Pan",
	"Expression",
	"Controller 12",
	"Controller 13",
	"Controller 14",
	"Controller 15",
	"Gen purp 1",
	"Gen purp 2",
	"Gen purp 3",
	"Gen purp 4",
	"Controller 20",
	"Controller 21",
	"Controller 22",
	"Controller 23",
	"Controller 24",
	"Controller 25",
	"Controller 26",
	"Controller 27",
	"Controller 28",
	"Controller 29",
	"Controller 30",
	"Controller 31",
	"Bank Sel LSB",
	"Modulation LSB",
	"Breath LSB",
	"Controller 35",
	"Foot LSB",
	"Portamento LSB",
	"Data Ent LSB",
	"Volume LSB",
	"Balance LSB",
	"Controller 41",
	"Pan LSB",
	"Expression LSB",
	"Controller 44",
	"Controller 45",
	"Controller 46",
	"Controller 47",
	"Controller 48",
	"Controller 49",
	"Controller 50",
	"Controller 51",
	"Controller 52",
	"Controller 53",
	"Controller 54",
	"Controller 55",
	"Controller 56",
	"Controller 57",
	"Controller 58",
	"Controller 59",
	"Controller 60",
	"Controller 61",
	"Controller 62",
	"Controller 63",
	"Sustain",
	"Porta on/off",
	"Sostenuto",
	"Soft Pedal",
	"Legato FS",
	"Hold 2",
	"Sound Var",
	"Harmonic",
	"Release Time",
	"Attack Time",
	"Brightness",
	"Controller 75",
	"Controller 76",
	"Controller 77",
	"Controller 78",
	"Controller 79",
	"Gen Purp 5",
	"Gen Purp 6",
	"Gen Purp 7",
	"Gen Purp 8",
	"Porta Ctrl",
	"Controller 85",
	"Controller 86",
	"Controller 87",
	"Controller 88",
	"Controller 89",
	"Controller 90",
	"ExtEff 1 Depth",
	"ExtEff 2 Depth",
	"ExtEff 3 Depth",
	"ExtEff 4 Depth",
	"ExtEff 5 Depth",
	"Data Incr",
	"Data Decr",
	"NRPN LSB",
	"NRPN MSB",
	"RPN LSB",
	"RPN MSB",
	"Controller 102",
	"Controller 103",
	"Controller 104",
	"Controller 105",
	"Controller 106",
	"Controller 107",
	"Controller 108",
	"Controller 109",
	"Controller 110",
	"Controller 111",
	"Controller 112",
	"Controller 113",
	"Controller 114",
	"Controller 115",
	"Controller 116",
	"Controller 117",
	"Controller 118",
	"Controller 119",
	"All Snd Off",
	"Reset Ctrl",
	"Local Ctrl",
	"AllNoteOff",
	"Omni Mode Off",
	"Omni Mode On",
	"Mono Mode On",
	"Poly Mode On",
};


