// Key mapping class

#include "stdafx.h"
#include "keymap.h"
#include "ctrlmap.h"
#include "midifish.h"



// construct a null key map
KeyMap::KeyMap(int key)
{
	int i, j;

	mKey = key;

	mSplitController[0] = CtrlMap::VELOCITY;
	mSplitController[1] = CtrlMap::CONTROLLER + 1;	// modulation

	for (i=0; i<2; i++) {
		mSplitCrossfade[i] = 0;
		mSplitPoint[i][0] = 0;

		for (j=1; j<=MAX_SPLITS; j++)
			mSplitPoint[i][j] = 256;
	}

	for (i=0; i<MAX_SPLITS; i++) {
		for (j=0; j<MAX_SPLITS; j++) {
			mMap[i][j].mKey = key;
			mMap[i][j].mPort = 0;
			mMap[i][j].mVelScale = 256;
			mMap[i][j].mVelOffset = 0;
		}
	}
}



// read a key mapping
void KeyMap::load(FILE *file)
{
	int i, j;

	for (i=0; i<2; i++) {
		mSplitController[i] = fgetl(file);
		mSplitCrossfade[i] = fgetl(file);

		for (j=1; j<MAX_SPLITS; j++)
			mSplitPoint[i][j] = fgetl(file);
	}

	for (i=0; i<MAX_SPLITS; i++) {
		for (j=0; j<MAX_SPLITS; j++) {
			mMap[i][j].mKey = fgetl(file);
			mMap[i][j].mPort = fgetl(file);
			mMap[i][j].mVelScale = fgetl(file);
			mMap[i][j].mVelOffset = fgetl(file);
		}
	}
}



// save a key mapping
void KeyMap::save(FILE *file)
{
	int i, j;

	for (i=0; i<2; i++) {
		fputl(mSplitController[i], file);
		fputl(mSplitCrossfade[i],  file);

		for (j=1; j<MAX_SPLITS; j++)
			fputl(mSplitPoint[i][j], file);
	}

	for (i=0; i<MAX_SPLITS; i++) {
		for (j=0; j<MAX_SPLITS; j++) {
			 fputl(mMap[i][j].mKey,       file);
			 fputl(mMap[i][j].mPort,      file);
			 fputl(mMap[i][j].mVelScale,  file);
			 fputl(mMap[i][j].mVelOffset, file);
		}
	}
}



// converts MIDI notes into string names
const char *KeyMap::note(int key)
{
	static const char *names[] = { "C", "C#", "D", "Eb", "E", "F", "F#", "G", "G#", "A", "Bb", "B" };
	
	return names[key % 12];
}



// converts MIDI notes into octave numbers
int KeyMap::octave(int key)
{
	return (key / 12) - 2;
}



// format the name of a single key
string KeyMap::get_name(int key)
{
	if (key < 0)
		return "<none>";

	char tmp[64];

	sprintf(tmp, "%s%d", note(key), octave(key));

	return tmp;
}



// get a string we can show to the user
string KeyMap::get_desc()
{
	char tmp[4096];

	sprintf(tmp, "%s%d\t-> ", note(mKey), octave(mKey));

	set<pair<int, int> > done;

	for (int i=0; i<MAX_SPLITS; i++) {
		if (mSplitPoint[0][i] >= 256)
			break;

		for (int j=0; j<MAX_SPLITS; j++) {
			if (mSplitPoint[1][j] >= 256)
				break;

			int key = mMap[i][j].mKey;
			int port = mMap[i][j].mPort;

			if ((key < 0) || (port < 0))
				continue;

			if (done.find(pair<int, int> (key, port)) != done.end())
				continue;

			if (!done.empty())
				strcat(tmp, "  +  ");

			sprintf(tmp+strlen(tmp), "%s%d", note(key), octave(key));

			if (port > 0)
				sprintf(tmp+strlen(tmp), " [%d]", port+1);

			done.insert(pair<int, int> (key, port));
		}
	}

	if (done.empty())
		strcat(tmp, "<none>");

	return tmp;
}



